# This is a wrapper around the Tensorflow.Model class. Here we can just add some meta data to keep things tidy
# I'm not implementing this yet because I think we won't be needing it. But I wanted this file here as
# a reminder.
import warnings

class CVModel:
    def getModel(self, *args):
        warnings.warn('WARNING: The getModel() function in class `{className}` has not been implemented.'.format(className = self.__class__))
        pass


import tensorflow as tf
from tensorflow.keras import metrics
def get_max_indices(heatmaps):
    batch_size = tf.shape(heatmaps)[0]
    hm_size = tf.cast(tf.shape(heatmaps)[2], tf.float32)
    num_joints = tf.shape(heatmaps)[3]
    tf.debugging.assert_equal(
        tf.rank(heatmaps), 
        4, 
        message='heatmaps should be of rank 4',
    )

    hm_reshaped = tf.reshape(heatmaps, [batch_size, -1, num_joints])
    max_vals = tf.reduce_max(hm_reshaped, axis=1)
    max_vals = tf.reshape(max_vals, [batch_size, 1, num_joints])

    max_idx = tf.argmax(hm_reshaped, axis=1)
    max_idx = tf.reshape(max_idx, [batch_size, 1, num_joints])
    max_idx = tf.cast(
        tf.tile(
            max_idx + 1, 
            multiples=[1, 2, 1],
        ), 
        tf.float32,
    )

    max_indices = tf.TensorArray(
        tf.float32, 
        size=2, 
        dynamic_size=False,
    )
    max_indices = max_indices.write(
        index=0, 
        value=(max_idx[:, 0, :] - 1) % hm_size + 1,
    )
    max_indices = max_indices.write(
        index=1, 
        value=tf.floor((max_idx[:, 1, :] - 1) / hm_size) + 1,
    )
    max_indices = max_indices.stack()
    max_indices = tf.transpose(max_indices, perm=[1,0,2])

    mask = tf.cast(
        tf.greater(max_vals, 0.0), 
        tf.float32,
    )
    mask = tf.tile(mask, multiples=[1,2,1])
    return max_indices * mask


NUM_JOINTS = 16
EXCLUDE = 6
PRECISION = 0.1
batch_size = 24
num_stacks = 4
hm_size = 64

class PercentageOfCorrectKeypoints(metrics.Metric):
    def __init__(self, name='pck', **kwargs):
        super(PercentageOfCorrectKeypoints, self).__init__(name=name)
        self.num_stacks = num_stacks
        self.hm_size = 64
        self.threshold = 0.5
        self.num_joints = NUM_JOINTS - EXCLUDE
        self.accuracy = tf.Variable(0.0, trainable=False)
        self.num_batches = tf.Variable(0.0, trainable=False)

    @classmethod
    def from_config(cls, config):
        return cls(**config)
    
    def get_config(self):
        config = {
            'num_stacks':self.num_stacks,
            'hm_size':self.hm_size,
            'threshold':self.threshold,
            'num_joints':self.num_joints,
        }
        base_config = super(PercentageOfCorrectKeypoints, self).get_config()    
        return dict(
            list(base_config.items()) + list(config.items())
        )

    @tf.function
    def update_state(
            self, 
            y_true, 
            y_pred, 
            sample_weight=None,
        ):
        tf.debugging.assert_equal(
            sample_weight, 
            None,
            message='sample_weight is not required',
        )
        hm_pred = tf.concat(
            values=[
                y_pred[...,:6], 
                y_pred[...,10:12], 
                y_pred[...,14:],
            ],
            axis=4,
        )
        hm_pred = hm_pred[:, self.num_stacks-1, :, :, :]
        hm_true = tf.concat(
            values=[
                y_true[...,:6], 
                y_true[...,10:12], 
                y_true[...,14:],
            ], 
            axis=3,
        )

        batch_acc = self._get_batch_acc(hm_pred, hm_true)
        self.accuracy.assign_add(batch_acc)
        self.num_batches.assign_add(1.)

    def result(self):
        return self.accuracy / self.num_batches

    def reset_state(self):
        self.accuracy.assign(0.)
        self.num_batches.assign(0.)

    def _get_batch_acc(self, hm_pred, hm_true):
        pred_max = get_max_indices(hm_pred)
        true_max = get_max_indices(hm_true)
        l2_distances = self._get_l2_distances(pred_max, true_max)

        batch_acc = tf.zeros([])
        visible_joints = tf.zeros([])

        joint_acc = tf.TensorArray(
            tf.float32, 
            size=self.num_joints+1, 
            dynamic_size=False, 
            clear_after_read=False,
        )
        for j in tf.range(self.num_joints):
            joint_acc = joint_acc.write(
                index=j, 
                value=self._get_joint_acc(l2_distances[:,j]),
            )
            if joint_acc.read(j) >= 0.:
                batch_acc = batch_acc + joint_acc.read(j)
                visible_joints = visible_joints + 1.

        if visible_joints != 0.:
            joint_acc = joint_acc.write(
                index=self.num_joints, 
                value=batch_acc / visible_joints,
            )
            joint_acc = joint_acc.stack()
            return joint_acc[-1]
        else:
            return tf.zeros([])

    def _get_l2_distances(self, pred_max, true_max):
        batch_size = tf.shape(true_max)[0]
        distances = tf.TensorArray(
            tf.float32, 
            size=batch_size, 
            dynamic_size=False,
        )
        for n in tf.range(batch_size):
            dist = tf.TensorArray(
                tf.float32, 
                size=self.num_joints, 
                dynamic_size=False,
            )
            for j in tf.range(self.num_joints):
                if true_max[n, 0, j] > 1 and true_max[n, 1, j] > 1:
                    dist = dist.write(
                        index=j, 
                        value=(
                            tf.norm(pred_max[n, :, j] - true_max[n, :, j]) 
                            / (self.hm_size * PRECISION)
                        )
                    )
                else:
                    dist = dist.write(j, -1.)
            dist = dist.stack()
            distances = distances.write(n, dist)
        return distances.stack()

    def _get_joint_acc(self, joint_dists):
        visible = joint_dists[tf.not_equal(joint_dists, -1.)]
        if tf.size(visible) > 0:
            below_thr = tf.less(visible, self.threshold)
            below_thr = tf.reduce_sum(
                tf.cast(below_thr, tf.float32)
            )
            return  below_thr / tf.cast(tf.size(visible), tf.float32)
        else:
            return -tf.ones([])
        

















# This is a wrapper around the functions that will be passed into map
import warnings

class Preprocessor:
    def getMap(self, *args):
        warnings.warn('WARNING: The getMap() function in class `{className}` has not been implemented.'.format(className = self.__class__))


#may take parameters, here it takes none
from Preprocessors.Preprocessor import Preprocessor
import numpy as np
import tensorflow as tf
import warnings

class OcclusionPreprocessor(Preprocessor):
    def getMap():
        def occlude_joints(im, j, wt, c, s):
            rng = tf.random.Generator.from_seed(1)
            indices_shuffled = tf.random.shuffle(tf.where(wt==1))[:,0]
            min_box_half_size, max_box_half_size, max_occlusions = 20, 30, len(indices_shuffled) + 1
            num_to_occlude = rng.uniform(shape=[], minval=0, maxval=max_occlusions, dtype=tf.int32)
            patch_half_width = rng.uniform(shape=[], minval=min_box_half_size, maxval=max_box_half_size, dtype=tf.int32)
            im_n = im.numpy()
            shape_x, shape_y = tf.cast(im_n.shape[1], tf.int32), tf.cast(im_n.shape[0], tf.int32)
            for i in range(num_to_occlude):
                index_to_occlude = indices_shuffled[i]
                x, y = tf.cast(j[index_to_occlude,0], tf.int32), tf.cast(j[index_to_occlude,1], tf.int32)
                min_y, max_y = max(y - patch_half_width, 0), min(y + patch_half_width + 1, shape_y)
                min_x, max_x = max(x - patch_half_width, 0), min(x + patch_half_width + 1, shape_x)
                im_n[min_y: max_y, min_x: max_x, :] = _get_background_patch(im, max_x - min_x, max_y - min_y)
                wt = tf.tensor_scatter_nd_update(wt, [[index_to_occlude]], [0])
            return im_n, j, wt
        # function which gets a patch of size height x width to occlude the image with 
        def _get_background_patch(im, width, height):
            return im[0:height, 0: width,:]
        def occlusion_wrapper(i, j, wt, loc):
            c = loc[0:2]
            s = loc[2]
            im, j, wt =  tf.py_function(occlude_joints, [i, j, wt, c, s], [tf.float32, tf.float32, tf.float32])
            im.set_shape(i.get_shape())
            return im, j, wt, loc
        return occlusion_wrapper


import tensorflow as tf
import Manifest
import pandas as pd 
import numpy as np

# the same thing in concept to getPreProcessor, returns tensorflow model
# every model will have an assocossiated ID, returns a tensorflow.model object 
def getModel(modelName: str):
    if (modelName not in Manifest.ModelIDs.keys()):
        raise KeyError
    cvModel = Manifest.ModelIDs[modelName]
    return cvModel.getModel()

# (csv location, folderOfImages location)
def getDataset(csvLocation: str, imagesLocation: str, datasetSpecName: str):
    df = pd.read_csv(csvLocation)
    imageNames = np.array([row[1].tolist()[0] for row in df.iterrows()])
    imageLabels = np.array([row[1].tolist()[1:] for row in df.iterrows()])
    datasetSpec = getDatasetSpec(datasetSpecName=datasetSpecName)
    dataset = tf.data.Dataset.from_tensor_slices((imageNames, imageLabels))
    #datasetSpec = getDatasetSpec(datasetSpecName)("","","")
    #image_label_pairs = np.array([datasetSpec.get_components(row[1].tolist()) for row in df.iterrows()])

    #imagesNames = image_label_pairs[:,0].tolist()
    #imageLabels = image_label_pairs[:,1].tolist()
    #raggedBoi = tf.ragged.constant((imagesNames, imageLabels))
    #ds_train = tf.data.Dataset.from_tensor_slices((imagesNames, imageLabels))
    def string_length_tf(t):
        return tf.py_function(len, [t], [tf.int64])
    def parse_image(filename, label):
        image_path = imagesLocation+"/"+filename
    #     image = tf.io.read_file(imagesLocation+"/"+filename)
    #    # try:
    #    #     # keep in mind that this assumes mac file system
    #    #     image = tf.io.read_file(imagesLocation+filename)
    #    # except:
    #    #     return np.zeros((10)), label
    #     image = tf.image.decode_image(image, channels=1, dtype=tf.float32)
    #     image.set_shape(image.get_shape())
    #     return image, label
        path_length = tf.strings.length(image_path)
        file_extension = tf.strings.substr(image_path, path_length - 3, 3)
        file_cond = tf.equal(file_extension, 'jpg')
        
        image  = tf.cond(file_cond, lambda: tf.image.decode_jpeg(tf.io.read_file(image_path)), lambda: tf.image.decode_png(tf.io.read_file(image_path)))
        return tf.cast(image, tf.float32) / 256.0, label
    def get_elements_wrapper(image, label):
       j, w, loc =  tf.py_function(datasetSpec.get_components, [label], [tf.float32, tf.float32, tf.float32])
       return image, j, w, loc

    dataset = dataset.map(parse_image) 
    return dataset.map(get_elements_wrapper)

# ds (dataset object)(tensorflow dataset)
# ds.map(function)
def getPreProcessor(preprocessorName: str):
    if (preprocessorName not in Manifest.PreprocessorIDs.keys()):
        raise KeyError
    preprocessor = Manifest.PreprocessorIDs[preprocessorName]
    return preprocessor.getMap()

def getDatasetSpec(datasetSpecName: str):
    if (datasetSpecName not in Manifest.DatasetSpecs.keys()):
        raise KeyError
    datasetSpec = Manifest.DatasetSpecs[datasetSpecName]
    return datasetSpec